/***************************************************************************
 main.cpp  -  description
 -------------------
 begin                : Fri Jan 17 17:59:14 UTC 2003
 copyright            : (C) 2003 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include "BioSequence.h"
#include "functions.h"
#include "cdb.h"
#include "badfilename_exception.h"
#include <map>
using namespace std;

/**
 * extracts a fragment xLength long starting at (xDist+xLength) left from given fragment [iDBF, iDBT]
 */
bool extractLeftFlank(CDb &ddb, const char *sDBL, BioSubSequence &sDB, int iDBF, int iDBT, int xDist, int xLength, bool bCountGap)
{
   sDB.clear();
   if (xLength)
   {
      xLength += ((xLength > 0) ? -1 : 1);
      int flankEnd = (iDBF <= iDBT) ? (iDBF - xDist) : (iDBF + xDist);
      int flankStart = (iDBF <= iDBT) ? (iDBF - xDist - xLength) : (iDBF + xDist + xLength);
      if (ddb.fragment(sDBL, flankStart, flankEnd, sDB, bCountGap) == 0)
      {
         cerr << sDBL << "\t" << flankStart << "\t" << flankEnd << "\t - not extracted\n";
      }
   }
   return (sDB.size() > 0);
}

/**
 * extracts a fragment xLength long starting at xDist right from given fragment [iDBF, iDBT]
 */
bool extractRightFlank(CDb &ddb, const char *sDBL, BioSubSequence &sDB, int iDBF, int iDBT, int xDist, int xLength, bool bCountGap)
{
   sDB.clear();
   if (xLength)
   {
      xLength += ((xLength > 0) ? -1 : 1);
      int flankStart = (iDBF <= iDBT) ? (iDBT + xDist) : (iDBT - xDist);
      int flankEnd = (iDBF <= iDBT) ? (iDBT + xDist + xLength) : (iDBT - xDist - xLength);
      if (ddb.fragment(sDBL, flankStart, flankEnd, sDB, bCountGap) == 0)
      {
         cerr << sDBL << "\t" << flankStart << "\t" << flankEnd << "\t - not extracted\n";
      }
   }
   return (sDB.size() > 0);
}
/*
 bool concatFlanks(CDb &ddb, const char *sDBL, BioSubSequence &sDB, int iDBF, int iDBT, int EXPANDL, int EXPANDR, bool bCountGap)
 {
 BioSubSequence seq;
 seq.clear();
 sDB.clear();
 if (EXPANDL)
 {
 extractLeftFlank(ddb, sDBL, sDB, iDBF, iDBT, 0, EXPANDL, bCountGap);
 }
 if (sDB.size() == 0)
 {
 sDB.setFrom((iDBF <= iDBT) ? (iDBT + 1) : (iDBT - 1));
 }
 if (EXPANDR)
 {
 extractRightFlank(ddb, sDBL, seq, iDBF, iDBT, 0, EXPANDR, bCountGap);
 sDB.append(seq);
 sDB.setTo(seq.getTo());
 }
 return (sDB.size() > 0);
 }
 */
/** ======================================================== */
void usage(char* exec)
{
   if (!exec)
   {
      return;
   }
   if (strrchr(exec, '/'))
   {
      exec = strrchr(exec, '/') + 1;
   }
   cerr << "NAME\n\t" << exec;
#ifdef VERSION

   cerr << " V" << VERSION;
#endif

   cerr << " - extracts corresponding fragments\n";
   cerr << "\nSYNOPSIS\n";
   cerr << "\t" << exec << " inmap(-) db_or_index_file\n";
   cerr << "\t[-xr #][-xl #][-fl #,#][-fr #,#][-cat0..3][-i1 -i2][-gap] [outfile]\n";
   cerr << "\nDESCRIPTION\n";
   cerr << "\tExtracts corresponding fragments from 'db', \n";
   cerr << "\twith optional expansion.\n";
   cerr << "\nOPTIONS\n";
   cerr << "\t-xl # (-xr #) - width of expansion flanks.\n";
   cerr << "\t-fl #,# (-fr #,#) - extract adjacent left (right) fragment\n";
   cerr << "\tdefined by a pair {distance from left (right) flank, length of the fragment}.\n";
   cerr << "\nFLAGS\n";
   cerr << "\t-cat0 - don't concatenate flanking regions, excluding middle from the output.\n";
   cerr << "\t-cat1 - don't concatenate flanking regions, including middle portion in the output.\n";
   cerr << "\t-cat2 - concatenate flanking regions, excluding middle from the output.\n";
   cerr << "\t-cat3 - concatenate flanking regions and middle portion. This is on by default.\n";
   cerr << "\t-c - complements extracted sequences.\n";
   cerr << "\t-gap - includes all nonblank symbols in calculating length.\n";
   cerr << "\t-i1 - includes information about position into sequence names.\n";
   cerr << "\t-i2 - includes information about position and corresponding\n";
   cerr << "\tmatching sequence from .map file into sequence names.\n";
   cerr << "\t-u - gives unique name to each extracted fragment.\n";
#ifdef PACKAGE_BUGREPORT

   cerr << "\nBUG REPORT\n\t" << PACKAGE_BUGREPORT << "\n";
#endif
#ifdef COPYRIGHT

   cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif
}
/** ======================================================== */
int main(int argc, char *argv[])
{
   istream *is;
   ostream *os;

   map<string, int> umap;
   ARGMAP amap;
   int nameInfoLevel = 0;
   int concatMode = 3;
   bool bCountGap = 0;
   bool bComplement = false;
   bool bCombineFlanks = false;
   int flDist = 0, flLength = 0;
   int frDist = 0, frLength = 0;
   string sLine;
   char sDBName[MAXNAMESIZE];
   char sQUName[MAXNAMESIZE];
   ::getoptions(argv, amap, 4, "-xl", "-xr", "-fl", "-fr");
   ::getflags(argv, amap, 11, "--help", "--version", "-u", "-i1", "-i2", "-gap", "-c", "-cat0", "-cat1", "-cat2", "-cat3");
   if (amap.find("--help") != amap.end() || amap.find("--version") != amap.end())
   {
      usage(argv[0]);
      exit(0);
   }
   if (amap.find("-u") != amap.end())
   {
      nameInfoLevel = 1;
   }
   if (amap.find("-i1") != amap.end())
   {
      nameInfoLevel = 2;
   }
   if (amap.find("-i2") != amap.end())
   {
      nameInfoLevel = 3;
   }
   if (amap.find("-cat0") != amap.end())
   {
      concatMode = 0;
   }
   if (amap.find("-cat1") != amap.end())
   {
      concatMode = 1;
   }
   if (amap.find("-cat2") != amap.end())
   {
      concatMode = 2;
   }
   if (amap.find("-cat3") != amap.end())
   {
      concatMode = 3;
   }
   if (amap.find("-gap") != amap.end())
   {
      bCountGap = true;
   }
   if (amap.find("-c") != amap.end())
   {
      bComplement = true;
   }
   if (amap.find("-cf") != amap.end())
   {
      cerr << "Received an -cf switch\n";
      bCombineFlanks = true;
   }
   if (amap.find("-xl") != amap.end())
   {
      if (sscanf(amap["-xl"].c_str(), "%d", &flLength) < 1)
      {
         fprintf(stderr, "Bad argument '%s' to -xl option\n", amap["-xl"].c_str());
         usage(argv[0]);
         exit(10);
      }
      flDist = 1;
   }
   else
   {
      if (amap.find("-fl") != amap.end())
      {
         if (sscanf(amap["-fl"].c_str(), "%d,%d", &flDist, &flLength) < 2)
         {
            fprintf(stderr, "Bad argument '%s' to -fl option\n", amap["-fl"].c_str());
            usage(argv[0]);
            exit(10);
         }
      }
   }
   if (amap.find("-xr") != amap.end())
   {
      if (sscanf(amap["-xr"].c_str(), "%d", &frLength) < 1)
      {
         fprintf(stderr, "Bad argument '%s' to -xr option\n", amap["-xr"].c_str());
         usage(argv[0]);
         exit(10);
      }
      frDist = 1;
   }
   else
   {
      if (amap.find("-fr") != amap.end())
      {
         if (sscanf(amap["-fr"].c_str(), "%d,%d", &frDist, &frLength) < 2)
         {
            fprintf(stderr, "Bad argument '%s' to -fr option\n", amap["-fr"].c_str());
            usage(argv[0]);
            exit(10);
         }
      }
   }
   char **args = argv;
   if ((args = assigninput(is, args)) == 0)
   {
      usage(argv[0]);
      exit(10);
   }
   if (args[1] == 0)
   {
      usage(argv[0]);
      exit(20);
   }
   strcpy(sDBName, args[1]);
   args++;
   if ((args = assignoutput(os, args)) == 0)
   {
      usage(argv[0]);
      exit(40);
   }
   //========================================================
   int iCnt = 0;
   try
   {
      const char FRAGMENT[] = "FRAGMENT";
      CDb ddb(sDBName);
      BioSubSequence sDB;
      int lineCount = 0;
      BioSubSequence seqFL, seqMID, seqFR;
      while (!getline(*is, sLine).fail())
      {
         lineCount++;
         long iDBF = 0, iDBT = 0;
         char sDBL[MAXLINELENGTH];
         sDBL[0] = 0;
         char sQUL[MAXNAMESIZE];
         sQUL[0] = '\0';
         long iQUF = 0, iQUT = 0;
         int colCnt = 0;
         if ((colCnt = sscanf(sLine.c_str(), "%s%ld%ld%s%ld%ld", sDBL, &iDBF, &iDBT, sQUL, &iQUF, &iQUT)) < 3)
         {
            const char *cptr;
            switch (sLine[0])
            {
               case '>':
                  sscanf(sLine.c_str() + 1, "%s", sDBL);
                  if ((cptr = strstr(sLine.c_str(), FRAGMENT)) != 0)
                  {
                     if (sscanf(cptr + strlen(FRAGMENT), "%ld -> %ld", &iDBF, &iDBT) < 2)
                     {
                        cerr << "Bad map row #" << lineCount << " :\n" << sLine;
                     }
                  }
                  break;
               case ';':
                  do
                  {
                     if ((cptr = strstr(sLine.c_str(), FRAGMENT)) != 0)
                     {
                        if (sscanf(cptr + strlen(FRAGMENT), "%ld -> %ld", &iDBF, &iDBT) < 2)
                        {
                           cerr << "Bad map row #" << lineCount << " :\n" << sLine;
                        }
                     }
                     sLine = "";
                  } while (!getline(*is, sLine).fail() && sLine[0] == ';');
                  sscanf(sLine.c_str(), "%s", sDBL);
                  break;
               default:
                  continue;
            }
            //          cerr << "Bad map row: \n" << sLine;
            if (iDBF == 0 && iDBT == 0)
            {
               continue;
            }
         }
         seqFL.clear();
         seqMID.clear();
         seqFR.clear();
         if (ddb.fragment(sDBL, iDBF, iDBT, seqMID, bCountGap) == 0)
         {
            cerr << sDBL << "\t" << iDBF << "\t" << iDBT << "\t - not extracted\n";
         }
         //cerr << seqMID.getName() << " concatMode=" << concatMode << endl;

         if (flLength || frLength)
         {
            if (flLength)
            {
               extractLeftFlank(ddb, sDBL, seqFL, iDBF, iDBT, flDist, flLength, bCountGap);
            }
            if (frLength)
            {
               extractRightFlank(ddb, sDBL, seqFR, iDBF, iDBT, frDist, frLength, bCountGap);
            }
         }
         BioSubSequence* sptrs[3] =
         { &seqFL, &seqMID, &seqFR };
         int outputSize = 3;
         switch (concatMode)
         {
            case 0:
               sptrs[1] = sptrs[2];
               outputSize = 2;
               break;
            case 1:
               break;
            case 2:
               sptrs[0]->append(*sptrs[2]);
               outputSize = 1;
               break;
            case 3:
               (*(sptrs[0])).append(*sptrs[1]).append(*sptrs[2]);
               outputSize = 1;
               break;
            default:
               break;
         }
         for (int i = 0; i < outputSize; i++)
         {
            if (sptrs[i]->size())
            {
               char sName[MAXNAMESIZE];
               switch (nameInfoLevel)
               {
                  case 0:
                     break;
                  case 1:
                     umap[sptrs[i]->getName()]++;
                     sprintf(sName, "%s_%d", sptrs[i]->getName().c_str(), umap[sptrs[i]->getName()]);
                     sptrs[i]->setName(sName);
                     break;
                  case 3:
                     if (colCnt >= 4)
                     {
                        sprintf(sName, "%s_[%ld-%ld]_[%ld-%ld]_%s", sptrs[i]->getName().c_str(), iDBF, iDBT, sptrs[i]->getFrom(), sptrs[i]->getTo(), sQUL);
                        sptrs[i]->setName(sName);
                        break;
                     }
                  case 2:
                     sprintf(sName, "%s_[%ld-%ld]_[%ld-%ld]", sptrs[i]->getName().c_str(), iDBF, iDBT, sptrs[i]->getFrom(), sptrs[i]->getTo());
                     sptrs[i]->setName(sName);
                     break;
               }
            }
         }
         if (bComplement)
         {
            for (int i = 0; i < outputSize; i++)
            {
               if (sptrs[i]->size())
               {
                  sptrs[i]->complement();
               }
            }
         }
         for (int i = 0; i < outputSize; i++)
         {
            if (sptrs[i]->size())
            {
#ifdef FASTA
               sptrs[i]->printFASTA(*os);
#else

               sptrs[i]->printIG(*os);
#endif
               iCnt++;
            }
         }
      }
   }
   catch (badfilename_exception& exc)
   {
      cerr << argv[0] << " : main() : badfilename '" << exc.what() << "'\n";
      exit(50);
   }
   catch (exception& exc)
   {
      cerr << exc.what();
      exit(60);
   }
   catch (...)
   {
      cerr << argv[0] << " : main() : bad exception\n";
      exit(70);
   }
   //========================================================
   os->flush();
   if (iCnt > 0)
   {
      return 0;
   }
   else
   {
      return 1;
   }
}
